#include <cstdlib>
#include <ctime>
#include <iostream>
#include <vector>

#include "rlcsa.h"
#include "misc/utils.h"

#ifdef MULTITHREAD_SUPPORT
#include <omp.h>
#endif


using namespace CSA;


const int MAX_THREADS = 64;
const usint MAX_OCCURRENCES = 100000;



void
printUsage()
{
  std::cout << "Usage: rlcsa_test [options] base_name [patterns [threads]]" << std::endl;
  std::cout << "  -l   Locate the occurrences." << std::endl;
  std::cout << "  -p   Pattern file is in Pizza & Chili format." << std::endl;
  std::cout << "  -w   Write the total number of occurrences of each suffix into patterns.found." << std::endl;
  std::cout << std::endl;
}


int main(int argc, char** argv)
{
  std::cout << "RLCSA test" << std::endl;
  std::cout << std::endl;

  bool locate = false, pizza = false, write = false;
  char* base_name = 0;
  char* patterns_name = 0;
  #ifdef MULTITHREAD_SUPPORT
  sint threads = 1;
  #endif
  for(int i = 1; i < argc; i++)
  {
    if(argv[i][0] == '-')
    {
      switch(argv[i][1])
      {
        case 'l':
          locate = true; break;
        case 'p':
          pizza = true; break;
        case 'w':
          write = true; break;
        default:
          std::cout << "Invalid option: " << argv[i] << std::endl << std::endl;
          printUsage();
          return 1;
      }
    }
    else if(base_name == 0)
    {
      base_name = argv[i];
    }
    else if(patterns_name == 0)
    {
      patterns_name = argv[i];
    }
    #ifdef MULTITHREAD_SUPPORT
    else
    {
      threads = std::min(MAX_THREADS, std::max(atoi(argv[i]), 1));
    }
    #endif
  }
  if(base_name == 0) { printUsage(); return 2; }

  std::cout << "Options:";
  if(locate) { std::cout << " locate"; }
  if(pizza) { std::cout << " pizza"; }
  write &= locate;
  if(write) { std::cout << " write"; }
  std::cout << std::endl;
  std::cout << "Base name: " << base_name << std::endl;
  if(patterns_name != 0) { std::cout << "Patterns: " << patterns_name << std::endl; }
  #ifdef MULTITHREAD_SUPPORT
  std::cout << "Threads: " << threads << std::endl; 
  #endif
  std::cout << std::endl;

  const RLCSA rlcsa(base_name, false);
  if(!rlcsa.isOk()) { return 3; }
  rlcsa.printInfo();
  rlcsa.reportSize(true);
  if(patterns_name == 0) { return 0; }

  std::ifstream patterns(patterns_name, std::ios_base::binary);
  if(!patterns)
  {
    std::cerr << "Error opening pattern file!" << std::endl;
    return 4;
  }

  std::vector<std::string> rows;
  if(!pizza) { readRows(patterns, rows, true); }
  else       { readPizzaChili(patterns, rows); }
  if(rows.size() == 0)
  {
    std::cerr << "No patterns found!" << std::endl;
    return 5;
  }

  int* totals = 0;
  if(write)
  {
    totals = new int[rlcsa.getSize()];
    for(usint i = 0; i < rlcsa.getSize(); i++) { totals[i] = 0; }
  }

  usint total = 0, total_size = 0, ignored = 0, found = 0;
  sint i, n = rows.size(); usint j;
  pair_type result;
  usint* matches;
  usint occurrences;

  double start = readTimer();
  #ifdef MULTITHREAD_SUPPORT
  omp_set_num_threads(threads);
  #pragma omp parallel private(j, result, matches, occurrences)
  {
    #pragma omp for schedule(dynamic, 1)
    for(i = 0; i < n; i++)
    {
      result = rlcsa.count(rows[i]);
      occurrences = length(result);
      #pragma omp critical
      {
        if(!isEmpty(result))
        {
          found++;
          if(occurrences <= MAX_OCCURRENCES) { total += occurrences; }
          else { ignored++; }
        }
        total_size += rows[i].length();
      }
      if(locate && !isEmpty(result) && occurrences <= MAX_OCCURRENCES)
      {
        matches = rlcsa.locate(result);
        if(write)
        {
          #pragma omp critical
          {
            for(j = 0; j < occurrences; j++) { totals[matches[j]]++; }
          }
        }
        delete[] matches;
      }
    }
  }
  #else
  for(i = 0; i < n; i++)
  {
    result = rlcsa.count(rows[i]);
    total_size += rows[i].length();
    occurrences = length(result);
    if(!isEmpty(result)) { found++; total += occurrences; }
    if(locate)
    {
      matches = rlcsa.locate(result);
      if(write)
      {
        for(j = 0; j < occurrences; j++) { totals[matches[j]]++; }
      }
      delete[] matches;
    }
  }
  #endif

  double time = readTimer() - start;
  double megabytes = total_size / (double)MEGABYTE;

  std::cout << "Patterns:     " << n << " (" << (n / time) << " / sec)" << std::endl;
  std::cout << "Total size:   " << megabytes << " MB";
  if(!locate)
  {
    std::cout << " (" << (megabytes / time) << " MB/s)";
  }
  std::cout << std::endl;
  std::cout << "Found:        " << found << std::endl;
  std::cout << "Occurrences:  " << total;
  if(locate)
  {
    std::cout << " (" << (total / time) << " / sec)";
  }
  std::cout << std::endl;
  std::cout << "Time:         " << time << " seconds" << std::endl;
  std::cout << std::endl;
  if(ignored > 0)
  {
    std::cout << "Ignored " << ignored << " patterns with more than " << MAX_OCCURRENCES << " occurrences." << std::endl;
    std::cout << std::endl;
  }

  if(write)
  {
    std::string output_name(patterns_name); output_name += ".found";
    std::ofstream output(output_name.c_str(), std::ios_base::binary);
    if(!output)
    {
      std::cerr << "Cannot open output file (" << patterns_name << ")!" << std::endl;
    }
    else
    {
      output.write((char*)totals, rlcsa.getSize() * sizeof(int));
      output.close();
    }
    delete[] totals; totals = 0;
  }

  return 0;
}
