#include <cstdlib>
#include <ctime>
#include <fstream>
#include <iostream>
#include <vector>

#include "rlcsa.h"


using namespace CSA;


const parameter_type INITIAL_ADJUSTMENT = parameter_type("INITIAL_ADJUSTMENT", 0);


int main(int argc, char** argv)
{
  std::cout << "Sampler test" << std::endl;
  if(argc < 3)
  {
    std::cout << "Usage: sampler_test base_name weights [multiplier]" << std::endl;
    return 1;
  }


  // Determine index parameters.
  std::string base_name(argv[1]);
  std::string parameters_name = base_name + PARAMETERS_EXTENSION;
  Parameters parameters;
  parameters.set(RLCSA_BLOCK_SIZE);
  parameters.set(SAMPLE_RATE);
  parameters.read(parameters_name);
  parameters.set(SUPPORT_LOCATE.first, 1);
  parameters.set(SUPPORT_DISPLAY.first, 1);
  parameters.set(WEIGHTED_SAMPLES.first, 1);
  // parameters.print();


  // Handle command line parameters.
  std::cout << "Base name: " << base_name << std::endl;
  std::cout << "Weights file: " << argv[2] << std::endl;
  int multiplier = 10;
  if(argc >= 4) { multiplier = std::max(1, atoi(argv[3])); }
  std::cout << "Multiplier: " << multiplier << std::endl;

  std::ifstream weights_file(argv[2], std::ios_base::binary);
  if(!weights_file)
  {
    std::cerr << "Error opening weights file!" << std::endl;
    return 2;
  }

  usint n = fileSize(weights_file) / sizeof(int);
  int* weights = new int[n];
  weights_file.read((char*)weights, n * sizeof(int));
  weights_file.close();
  for(usint i = 0; i < n; i++)
  {
    if(weights[i] == 0) { weights[i] = 1; }
    else if(weights[i] > 0) { weights[i] *= multiplier; }
    else
    {
      std::cerr << "Negative value found: weights[" << i << "] == " << weights[i] << "!" << std::endl;
      delete[] weights;
      return 3;
    }
  }
  std::cout << "Input size: " << n << std::endl;
  usint samples = std::max((usint) 1, n / parameters.get(SAMPLE_RATE));
  std::cout << "Samples: " << samples << std::endl;
  std::cout << std::endl;


  // Do the actual sampling.
  double start = readTimer();
  WeightedSampler sampler(weights, n, true);
  sampler.buildSamples(samples, parameters.get(INITIAL_ADJUSTMENT));
  double sampled = readTimer();


  // Build RLCSA.
  std::ifstream input_file(base_name.c_str(), std::ios_base::binary);
  if(!input_file)
  {
    std::cerr << "Error opening input file " << base_name << "!" << std::endl;
    return 4;
  }
  uchar* data = new uchar[n];
  input_file.read((char*)data, n);
  input_file.close();
  RLCSA rlcsa(data, n, parameters.get(RLCSA_BLOCK_SIZE), sampler, true);
  if(rlcsa.isOk())
  {
    rlcsa.printInfo();
    rlcsa.reportSize(true);
    rlcsa.writeTo(base_name);
  }
  double stop = readTimer();


  double sampling_time = sampled - start;
  double time = stop - start;
  double megabytes = n / (double)MEGABYTE;
  std::cout << "Total: " << time << " seconds (" << (megabytes / time) << " MB/s)" << std::endl;
  std::cout << "Sampling: " << sampling_time << " seconds (" << (megabytes / sampling_time) << " MB/s)" << std::endl;
  std::cout << std::endl;

  return 0;
}
